from mitmproxy import http
from rikki.mitmproxy.model import Request, Response
from typing import List, Optional, Dict
from mitmproxy.coretypes import multidict
from mitmproxy.net.http import Headers
from deepdiff import DeepDiff
import json
import re


def is_json(content: str):
    result = None
    try:
        result = json.loads(content)
    except ValueError:
        pass
    return result


def is_html(content: str):
    if content.lower().startswith("<!DOCTYPE html>".lower()):
        return True
    else:
        return False


def content_type(content: str) -> str:
    if is_json(content):
        return "application/json; charset=UTF-8"
    elif is_html(content):
        return "text/html; charset=utf-8"
    else:
        return "text/plain"


def replace(flow: http.HTTPFlow, request: Optional[Request] = None, response: Optional[Response] = None):
    if request:
        if request.host:
            flow.request.host = request.host
        if request.content:
            flow.request.content = request.content.encode("utf8")
        if request.params:
            for item in request.params.items():
                flow.request.query.set_all(item[0], [item[1]])
        if request.headers:
            for item in request.headers.items():
                headers: Headers = flow.request.headers
                headers.set_all(item[0], [item[1]])
    if response:
        if response.headers:
            for item in response.headers.items():
                headers: Headers = flow.response.headers
                headers.set_all(item[0], [item[1]])
        if response.content:
            flow.response.content = response.content.encode("utf8")
            flow.response.headers["Content-Type"] = content_type(response.content)
        if response.code:
            flow.response.status_code = response.code


def filter(
        flows: List[http.HTTPFlow],
        request: Optional[Request] = None,
        response: Optional[Response] = None
) -> List[http.HTTPFlow]:
    """
    Filter flows and return only those that contain data from filter params
    :param flows: list of flows to be filtered
    :param request: request data to look for during filter
    :param response: response data to look for during filter
    :return: new list of filtered flows
    """
    result: List[http.HTTPFlow] = []
    result.extend(flows)

    for flow in result:
        if request:
            if request.port and request.port != flow.request.port:
                if flow in result:
                    result.remove(flow)
                    continue
            if request.path and request.path != flow.request.path.split("?")[0]:
                if flow in result:
                    result.remove(flow)
                    continue
            if request.method and request.method != flow.request.method:
                if flow in result:
                    result.remove(flow)
                    continue
            if request.host:
                host_pattern = re.compile(f"{request.host}")
                """
                Use `Host` header as fallback when proxy works as transparent proxy and has only ip addresses
                """
                request_host = flow.request.host
                header_host = ""
                host_header_name = 'Host'
                if host_header_name in flow.request.headers:
                    header_host = flow.request.headers[host_header_name]
                if not host_pattern.fullmatch(request_host) and not host_pattern.fullmatch(header_host):
                    if flow in result:
                        result.remove(flow)
            if request.params and not contains_params(flow.request.query, request.params):
                if flow in result:
                    result.remove(flow)
                    continue
            if request.headers and not contains_headers(flow.request.headers, request.headers):
                if flow in result:
                    result.remove(flow)
                    continue
            if request.content:
                try:
                    content_decode = flow.request.raw_content.decode("utf8")
                except:
                    result.remove(flow)
                    continue

                if not compare_content(request.content, content_decode):
                    if flow in result:
                        result.remove(flow)
                        continue
        if response:
            if response.code and response.code != flow.response.status_code:
                if flow in result:
                    result.remove(flow)
            if response.headers and not contains_headers(flow.response.headers, response.headers):
                if flow in result:
                    result.remove(flow)
            if response.content and flow.response.content:
                try:
                    content_decode = flow.response.content.decode("utf8")
                except:
                    result.remove(flow)
                    continue

                if response.content.lower().strip() != content_decode.lower().strip():
                    if flow in result:
                        result.remove(flow)

    return result


def contains_params(base: multidict.MultiDictView, candidate: Dict) -> bool:
    for key in candidate.keys():
        if key in base.keys() and candidate[key] in base.get_all(key):
            continue
        else:
            return False
    return True


def contains_headers(headers: Headers, candidate: Dict[str, str]) -> bool:
    for key in candidate.keys():
        if key in headers.keys() and candidate[key] in headers.get_all(key):
            continue
        else:
            return False
    return True


def compare_content(left, right) -> bool:
    """
    If the left and right are JSON, will check that all key and values from left are in right.
    If they are not JSON. will compare as lower cased strings.
    THE IMPLEMENTATION RIGHT NOW IS AWKWARD, BUT IT WORKS.
    TODO: it's better to revisit the implementation and simple allow users to find all key and values from left in right
    :param left:
    :param right:
    :return:
    """
    left_json = None
    right_json = None

    if isinstance(left, str) and isinstance(right, str):
        if left.startswith("?{") or left.startswith("?["):
            left_json = is_json(left[1:])
            right_json = is_json(right)
            if not left_json or not right_json:
                return left[1:].lower().strip() == right.lower().strip()
        else:
            return left.lower().strip() == right.lower().strip()
    else:
        left_json = left
        right_json = right

    if left_json and right_json:
        diff = DeepDiff(left_json, right_json, ignore_order=True)
        if not diff:
            return True
        else:
            """
            Check if any of item from left was removed. We don't care if something was added from left
            """
            added_items = []
            removed_items = []
            changed_items = []
            for item in diff.items():
                if str(item[0]).endswith("_removed"):
                    removed_items.append(item[1])
                elif str(item[0]).endswith("_added"):
                    added_items.append(item[1])
                elif str(item[0]).endswith("_changed"):
                    changed_items.append(item[1])
                else:
                    pass

            if changed_items:
                return False

            if removed_items:
                for removed_item in removed_items:
                    item_completely_removed = True
                    if added_items:
                        for added_item in added_items:
                            if "root['root" in str(removed_item):
                                break
                            if compare_content(removed_item, added_item):
                                item_completely_removed = False
                                break
                        if item_completely_removed:
                            return False  # return False
            return True



from typing import List
from unittest import TestCase

from mitmproxy import http

from rikki.mitmproxy.delgate.dump import DumpWriterConfig, DumpWriter, DumpWriterDelegate
from test.rikki.mitmproxy.utils import build_flow
from rikki.mitmproxy.model import Request, Response


class TestDump(TestCase):

    def test_should_dump_filtered_requests(self):
        flow_one = build_flow(request=Request(host="http://should_be_in_results"))
        flow_two = build_flow(request=Request(host="http://should_be_in_results"))
        flow_three = build_flow(request=Request(host="http://should_not_be_in_results"))
        flow_for = build_flow(request=Request(host="http://should_not_be_in_results"))

        config = DumpWriterConfig(
            path="",
            request=Request(host="http://should_be_in_results"),
            dump_writer=_MockDumpWriter
        )

        dump = DumpWriterDelegate(config=config)

        dump.request(flow_one)
        dump.request(flow_two)
        dump.request(flow_three)
        dump.request(flow_for)

        dump.shutdown()

        # noinspection PyTypeChecker
        dump_writer: _MockDumpWriter = dump.writer

        self.assertEqual(len(dump_writer.results), 2)
        self.assertIn(flow_one, dump_writer.results)
        self.assertIn(flow_two, dump_writer.results)

    def test_should_close_dump_writer(self):
        config = DumpWriterConfig(
            path="",
            dump_writer=_MockDumpWriter
        )

        dump = DumpWriterDelegate(config=config)

        dump.shutdown()

        # noinspection PyTypeChecker
        dump_writer: _MockDumpWriter = dump.writer

        self.assertTrue(dump_writer.closed)


class _MockDumpWriter(DumpWriter):

    def __init__(self, path: str) -> None:
        super().__init__(path)
        self.results: List[http.HTTPFlow] = []
        self.closed: bool = False

    def write_flows(self, flows: List[http.HTTPFlow]):
        self.results.extend(flows)

    def close(self):
        self.closed = True



from behave import given, then

from rikki.behave.context import Context
from rikki.behave.steps.utils import http_data
from rikki.behave.steps.utils import verify_number_of_request
from rikki.mitmproxy.delegate_replace import ReplacementConfig, InterceptAndReplace
from rikki.mitmproxy.delegate_delay import DelayConfig, Delay
from rikki.mitmproxy.delgate.dump import DumpWriterConfig, DumpWriterDelegate


@given("Delay request for {number:d} seconds")
def step_delay(context: Context, number):
    http_step_data = http_data(context)

    if http_step_data:
        config = DelayConfig(
            request=http_step_data.filter_request,
            response=http_step_data.filter_response,
            delay=number
        )
        plugin = Delay(config=config)
        context.proxy.add_delegate(plugin)


@given("Randomly delay request for {number:d} seconds")
def step_random_delay(context: Context, number):
    http_step_data = http_data(context)

    if http_step_data:
        config = DelayConfig(
            request=http_step_data.filter_request,
            response=http_step_data.filter_response,
            delay=number,
            random=True
        )
        plugin = Delay(config=config)
        context.proxy.add_delegate(plugin)


@given("Intercept and replace http data")
def step_replace_http_data(context: Context):
    http_step_data = http_data(context)

    if http_step_data:
        config = ReplacementConfig(
            filter_request=http_step_data.filter_request,
            filter_response=http_step_data.filter_response,
            replacement_request=http_step_data.replacement_request,
            replacement_response=http_step_data.replacement_response,
            wait_response=False
        )
        plugin = InterceptAndReplace(config=config, recursive=False)
        context.proxy.add_delegate(plugin)


@given("Replace http data")
def step_replace_http_data(context: Context):
    http_step_data = http_data(context)

    if http_step_data:
        config = ReplacementConfig(
            filter_request=http_step_data.filter_request,
            filter_response=http_step_data.filter_response,
            replacement_request=http_step_data.replacement_request,
            replacement_response=http_step_data.replacement_response,
        )
        plugin = InterceptAndReplace(config=config, recursive=False)
        context.proxy.add_delegate(plugin)


@then('Verify {number:d} requests for "{tag}"')
def step_verify_number_of_requests(context: Context, number, tag):
    verify_number_of_request(
        context=context,
        number=number,
        tag=tag,
        second_attempt=True
    )


@then('Check no "{data}" in any requests/responses because "{tag}"')
def step_check_for_data_in_http(context: Context, data: str, tag):
    verify_number_of_request(
        context=context,
        number=0,
        tag=tag,
        second_attempt=True,
        in_send=data,
        in_receive=data
    )


@then('Check no "{data}" in any request because "{tag}"')
def step_check_for_request_in_http(context: Context, data: str, tag):
    verify_number_of_request(
        context=context,
        number=0,
        tag=tag,
        second_attempt=True,
        in_send=data
    )


@then('Check no "{data}" in any responses because "{tag}"')
def step_check_for_response_in_http(context: Context, data: str, tag):
    verify_number_of_request(
        context=context,
        number=0,
        tag=tag,
        second_attempt=True,
        in_receive=data
    )


@given('Save network data in "{path}"')
def step_save_network_flows(context: Context, path: str):
    http_step_data = http_data(context)
    config = DumpWriterConfig(
        request=http_step_data.filter_request,
        response=http_step_data.filter_response,
        path=path
    )
    context.proxy.add_delegate(DumpWriterDelegate(config))



from typing import Dict, Optional


class Request:
    def __init__(
            self,
            host: Optional[str] = None,
            port: Optional[int] = None,
            path: Optional[str] = None,
            method: Optional[str] = None,
            params: Optional[Dict[str, object]] = None,
            headers: Optional[Dict[str, str]] = None,
            content: Optional[str] = None
    ) -> None:
        super().__init__()
        self.host: Optional[str] = host
        self.port: Optional[int] = port
        self.path: Optional[str] = path
        self.method: Optional[str] = method
        self.params: Optional[Dict[str, object]] = params
        self.headers: Optional[Dict[str, str]] = headers
        self.content: Optional[str] = content

    def empty(self) -> bool:
        if self.host or self.content or self.params or self.headers or self.port or self.path or self.method:
            return False
        else:
            return True


class Response:
    def __init__(
            self,
            headers: Optional[Dict[str, str]] = None,
            content: Optional[str] = None,
            code: Optional[int] = None
    ) -> None:
        super().__init__()
        self.headers: Optional[Dict[str, str]] = headers
        self.__content: Optional[str] = content
        self.__code: Optional[int] = code

    @property
    def code(self) -> int:
        return self.__code

    @code.setter
    def code(self, value):
        self.__code = int(value)

    @property
    def content(self) -> str:
        return self.__content

    @content.setter
    def content(self, value):
        self.__content = str(value)

    def empty(self) -> bool:
        if self.code or self.headers or self.content:
            return False
        else:
            return True



from behave import given

from rikki.behave.context import Context
import time


@given('Wait for shutdown')
def step_wait_for_shutdown(context: Context):
    time.sleep(100)
    step_wait_for_shutdown(context)



from typing import Optional

from selenium.webdriver.remote.webdriver import WebDriver

from rikki.appium.common import AppiumUtils, Direction, SwipeSpeed


class IOSXCUITestUtils(AppiumUtils):

    def __init__(self, driver: WebDriver, wait_time: int = 1) -> None:
        super().__init__(driver, wait_time)
        self._directions = {
            Direction.UP: "up",
            Direction.DOWN: "down",
            Direction.LEFT: "left",
            Direction.RIGHT: "right"
        }

    def relaunch_app(self):
        self._driver.launch_app()

    def swipe(self, element: Optional[object] = None, direction: Direction = Direction.DOWN,
              speed: SwipeSpeed = SwipeSpeed.MEDIUM):
        if element:
            self._driver.execute_script('mobile: scroll',
                                        {
                                            'direction': self._directions[direction],
                                            'element': f"{element}"
                                        }
                                        )
        else:
            self._driver.execute_script('mobile: scroll', {'direction': self._directions[direction]})

    @staticmethod
    def support(driver: WebDriver) -> bool:
        platform = driver.desired_capabilities['desired']['platformName']
        automation = driver.desired_capabilities['desired']['automationName']

        return platform == "iOS" and automation == "XCUITest"



from typing import List, IO, Optional, Type
from mitmproxy import io, http
from mitmproxy.exceptions import FlowReadException

from rikki.mitmproxy.plugin import ProxyDelegate
from rikki.mitmproxy.model import Request, Response
from rikki.mitmproxy.util import filter as filter_requests
from pathlib import Path


class DumpWriter:

    def __init__(self, path: str) -> None:
        self._path: Path = Path(path)

    def write_flows(self, flows: List[http.HTTPFlow]): assert False, f"{self} doesn't implement the method"

    def close(self): pass


class DumpReader:

    # noinspection PyTypeChecker
    def read_from(self, path: str) -> List[http.HTTPFlow]: assert False, f"{self} doesn't implement the method"


class MitmProxyDumpWriter(DumpWriter):

    def __init__(self, path: str) -> None:
        super().__init__(path)
        self.__file: IO[bytes] = open(str(self._path.absolute()), "wb")
        self.__writer = io.FlowWriter(self.__file)

    def write_flows(self, flows: List[http.HTTPFlow]):
        for flow in flows:
            self.__writer.add(flow)

    def close(self):
        self.__file.close()


class MitmProxyDumpReader(DumpReader):

    def read_from(self, path: str) -> List[http.HTTPFlow]:
        with open(path, "rb") as dump_file:
            flow_reader = io.FlowReader(dump_file)
            try:
                results: List[http.HTTPFlow] = []
                for flow in flow_reader.stream():
                    results.append(flow)
                return results
            except FlowReadException as e:
                assert False, f"File {path} corrupted"


class DumpWriterConfig:

    def __init__(
            self,
            path: str,
            dump_writer: Type[DumpWriter] = MitmProxyDumpWriter,
            request: Optional[Request] = None,
            response: Optional[Response] = None
    ) -> None:
        self.request: Optional[Request] = request
        self.response: Optional[Response] = response
        self.path: Optional[str] = path
        self.dump_writer: Type[DumpWriter] = dump_writer


class DumpWriterDelegate(ProxyDelegate):

    def __init__(self, config: DumpWriterConfig, recursive: bool = True) -> None:
        super().__init__(recursive=recursive)
        self.__config = config
        self.__flows: List[http.HTTPFlow] = []
        self.writer: DumpWriter = self.__config.dump_writer(path=self.__config.path)

    def request(self, flow: http.HTTPFlow):
        self.__flows.extend(
            filter_requests(
                flows=[flow],
                request=self.__config.request,
                response=self.__config.response
            )
        )

    def shutdown(self):
        self.writer.write_flows(self.__flows)
        self.writer.close()

    def clean_up(self):
        super().clean_up()



from unittest import TestCase
from test.rikki.mitmproxy.utils import build_flow
from rikki.mitmproxy.model import Request
from rikki.mitmproxy.util import filter
from json import JSONEncoder


class TestFlowFilter(TestCase):

    def test_plain_host(self):
        flow = build_flow(request=Request(host="www.google.com"))

        result = filter([flow], request=Request(host="google"))

        assert len(result) == 0

        result = filter([flow], request=Request(host=".*google"))

        assert len(result) == 0

        result = filter([flow], request=Request(host=".*google.*"))

        assert len(result) == 1

    def test_plain_host_and_host_header(self):
        flow = build_flow(request=Request(host="127.0.0.1", headers={"Host": "www.google.com"}))

        result = filter([flow], request=Request(host="google"))

        assert len(result) == 0

        result = filter([flow], request=Request(host=".*google"))

        assert len(result) == 0

        result = filter([flow], request=Request(host=".*google.*"))

        assert len(result) == 1

    def test_plain_content(self):
        flow = build_flow(
            request=Request(
                host="127.0.0.1",
                content="hello"
            )
        )

        result = filter(
            [flow],
            request=Request(
                host="127.0.0.1",
                content="no way"
            )
        )

        self.assertEqual(len(result), 0)

        result = filter(
            [flow],
            request=Request(
                host="127.0.0.1",
                content="hello"
            )
        )

        self.assertEqual(len(result), 1)

    def test_plain_host_and_json_content_json_dict(self):
        flow = build_flow(
            request=Request(
                host="127.0.0.1",
                content="{\"host\": \"www.google.com\", \"name\": \"rikki\"}"
            )
        )

        result = filter(
            [flow],
            request=Request(
                host="127.0.0.1",
                content="?{\"evil\": \"www.yandex.com\"}"
            )
        )

        self.assertEqual(len(result), 0)

        result = filter(
            [flow],
            request=Request(
                host="127.0.0.1",
                content="?{\"host\": \"www.google.com\"}"
            )
        )

        self.assertEqual(len(result), 1)

        result = filter(
            [flow],
            request=Request(
                host="127.0.0.1",
                content="?{\"host\": \"www.google.com\", \"name\": \"rikki\"}"
            )
        )

        self.assertEqual(len(result), 1)

    def test_plain_host_and_json_content_json_array(self):
        flow = build_flow(
            request=Request(
                host="127.0.0.1",
                content="[{\"host\": \"www.google.com\", \"name\": \"rikki\"}]"
            )
        )

        result = filter(
            [flow],
            request=Request(
                host="127.0.0.1",
                content="[{\"evil\": \"www.yandex.com\"}]"
            )
        )

        self.assertEqual(len(result), 0)

        result = filter(
            [flow],
            request=Request(
                host="127.0.0.1",
                content="[{\"host\": \"www.google2.com\"}]"
            )
        )

        self.assertEqual(len(result), 0)

        result = filter(
            [flow],
            request=Request(
                host="127.0.0.1",
                content="[{\"host\": \"www.google.com\", \"name\": \"rikki\"}]"
            )
        )

        self.assertEqual(len(result), 1)

    def test_nested_json_object_in_content(self):
        flow = build_flow(
            request=Request(
                host="127.0.0.1",
                content="{\"host\": \"www.google.com\", \"config\": { \"name\": \"rikki\", \"lang\": \"pyhton\" }}"
            )
        )

        result = filter(
            [flow],
            request=Request(
                host="127.0.0.1",
                content="?{\"config\": {\"name\": \"rikki\"}}"
            )
        )

        self.assertEqual(len(result), 1)

    def test_nested_object_in_array(self):
        flow = build_flow(
            request=Request(
                content=JSONEncoder().encode(
                    {
                        "param": "",
                        "configs": [
                            {
                                "name": "rikki",
                                "lang": "python",
                                "version": "unknown"
                            },
                            {
                                "name": "undefined",
                                "lang": "undefined"
                            }
                        ]
                    }
                )
            )
        )

        request = Request(
            content="?" + JSONEncoder().encode(
                {
                    "param": "",
                    "configs": [
                        {
                            "name": "rikki",
                            "lang": "python"
                        }
                    ]
                }
            )
        )

        result = filter([flow], request)

        self.assertEqual(len(result), 1)



import unittest

from behave.model import Table

from rikki.behave.context import Context
from rikki.behave.steps.utils import create_pattern, http_data


class TestUtils(unittest.TestCase):

    def test_should_create_simplified_pattern(self):
        pattern = create_pattern("*google*")

        self.assertNotEqual(pattern.fullmatch("www.google.com"), None,
                            "Simplified Wildcard pattern doesn't match value")

        pattern = create_pattern("*google.com")

        self.assertNotEqual(pattern.fullmatch("www.google.com"), None,
                            "Simplified Wildcard pattern doesn't match value")

        pattern = create_pattern("google.com")

        self.assertEqual(pattern.fullmatch("www.google.com"), None,
                         "Simplified Wildcard pattern match value, but shouldn't")

        pattern = create_pattern("127.0.0.1")

        self.assertNotEqual(pattern.fullmatch("127.0.0.1"), None,
                            "Simplified Wildcard pattern doesn't match value")

        self.assertEqual(pattern.fullmatch("127.0.0.1:8080"), None,
                         "Simplified Wildcard pattern match value, but shouldn't")

        pattern = create_pattern("*127.0.0.1*")

        self.assertEqual(pattern.fullmatch("127x0x0x1"), None,
                         "Simplified Wildcard pattern match value, but shouldn't")

    def test_should_create_full_featured_pattern(self):
        pattern = create_pattern("r\".*127.0.0.1.*\"")

        self.assertNotEqual(pattern.fullmatch("127.0.0.1"), None,
                            "Simplified Wildcard pattern doesn't match value")

        self.assertNotEqual(pattern.fullmatch("127x0x0x1"), None,
                            "Simplified Wildcard pattern doesn't match value")
        self.assertEqual(pattern.fullmatch("127..0.0.1"), None,
                         "Simplified Wildcard pattern match value, but shouldn't")

        pattern = create_pattern("r\".*127\.0\.0\.1\.*\"")

        self.assertNotEqual(pattern.fullmatch("127.0.0.1"), None,
                            "Simplified Wildcard pattern doesn't match value")

        self.assertEqual(pattern.fullmatch("127x0x0x1"), None,
                         "Simplified Wildcard pattern match value, but shouldn't")
        self.assertEqual(pattern.fullmatch("127..0.0.1"), None,
                         "Simplified Wildcard pattern match value, but shouldn't")

    def test_should_parse_http_data(self):
        table = Table(headings=["option", "value"])
        table.add_row(["port", "8080"])
        table.add_row(["host", "localhost"])
        table.add_row(["method", "GET"])
        table.add_row(["path", "/path/to/"])
        table.add_row(["headers", "{\"header1\": \"1\"}"])
        table.add_row(["params", "{\"param1\": \"1\"}"])
        table.add_row(["content", "content"])
        table.add_row(["response_code", "200"])
        table.add_row(["response_headers", "{\"response_header1\": \"1\"}"])
        table.add_row(["response_content", "response content"])
        table.add_row(["replace_port", "8181"])
        table.add_row(["replace_host", "127.0.0.1"])
        table.add_row(["replace_path", "/to/path/"])
        table.add_row(["replace_headers", "{\"replace_header1\": \"1\"}"])
        table.add_row(["replace_params", "{\"replace_param1\": \"1\"}"])
        table.add_row(["replace_content", "replace content"])
        table.add_row(["replace_response_code", "201"])
        table.add_row(["replace_response_content", "replace response content"])
        table.add_row(["replace_response_headers", "{\"replace_response_header1\": \"1\"}"])

        context = Context(table=table)

        result = http_data(context)

        self.assertEqual(result.filter_request.port, "8080")
        self.assertEqual(result.filter_request.method, "GET")
        self.assertEqual(result.filter_request.host, "localhost")
        self.assertEqual(result.filter_request.path, "/path/to/")
        self.assertEqual(result.filter_request.headers, {"header1": "1"})
        self.assertEqual(result.filter_request.params, {"param1": "1"})
        self.assertEqual(result.filter_request.content, "content")

        self.assertEqual(result.filter_response.code, 200)
        self.assertEqual(result.filter_response.headers, {"response_header1": "1"})
        self.assertEqual(result.filter_response.content, "response content")

        self.assertEqual(result.replacement_request.port, "8181")
        self.assertEqual(result.replacement_request.host, "127.0.0.1")
        self.assertEqual(result.replacement_request.path, "/to/path/")
        self.assertEqual(result.replacement_request.headers, {"replace_header1": "1"})
        self.assertEqual(result.replacement_request.params, {"replace_param1": "1"})
        self.assertEqual(result.replacement_request.content, "replace content")

        self.assertEqual(result.replacement_response.code, 201)
        self.assertEqual(result.replacement_response.headers, {"replace_response_header1": "1"})
        self.assertEqual(result.replacement_response.content, "replace response content")



from typing import Optional, List

from mitmproxy import http

from rikki.mitmproxy.model import Request, Response
from rikki.mitmproxy.plugin import ProxyDelegate
from rikki.mitmproxy.util import filter


class AccumulateFilterConfig:

    def __init__(
            self,
            request: Optional[Request] = None,
            response: Optional[Response] = None,
            in_send: Optional[str] = None,
            in_receive: Optional[str] = None
    ) -> None:
        super().__init__()
        self.request = request
        self.response = response
        self.in_send = in_send
        self.in_receive = in_receive


class Accumulate(ProxyDelegate):

    def __init__(self, filter_config: AccumulateFilterConfig):
        """
        This plugin will accumulate all flows filtered with the config. Be aware that flow will appear in results
        with delay in case when you specify response in filter config
        :param filter_config: request and response data, that will be used to filter requests
        """
        super().__init__()
        self.__config = filter_config
        self.accumulated_items: List[http.HTTPFlow] = []

    def request(self, flow: http.HTTPFlow):
        if self.__config.request and not self.__config.response:
            filtered_by_request = filter([flow], request=self.__config.request)
            if filtered_by_request and self.__config.in_send:
                if self.__check_everywhere(flow.request, self.__config.in_send):
                    self.accumulated_items.extend(filtered_by_request)
            else:
                self.accumulated_items.extend(filtered_by_request)
        if not self.__config.request and not self.__config.response:
            if self.__config.in_send and not self.__config.in_receive:
                if self.__check_everywhere(flow.request, self.__config.in_send):
                    self.accumulated_items.append(flow)

    def response(self, flow: http.HTTPFlow):
        if self.__config.response:
            filtered_flow = filter([flow], request=self.__config.request, response=self.__config.response)

            if filtered_flow and (self.__config.in_send or self.__config.in_receive):
                something_in_request = self.__check_everywhere(flow.request, self.__config.in_send)
                something_in_response = self.__check_everywhere(flow.request, self.__config.in_receive)
                if something_in_request or something_in_response:
                    self.accumulated_items.extend(filtered_flow)
            else:
                self.accumulated_items.extend(filtered_flow)

            if flow not in self.accumulated_items and \
                    self.__check_everywhere(data=flow.response, value=self.__config.in_receive):
                self.accumulated_items.append(flow)
        if not self.__config.request and not self.__config.response:
            if self.__config.in_send or self.__config.in_receive:
                something_in_request = self.__check_everywhere(flow.request, self.__config.in_send)
                something_in_response = self.__check_everywhere(flow.request, self.__config.in_receive)
                if something_in_request or something_in_response:
                    self.accumulated_items.append(flow)

    def clean_up(self):
        self.accumulated_items.clear()

    @staticmethod
    def __check_everywhere(data, value: Optional[str]) -> bool:
        if not value:
            return False
        if isinstance(data, http.HTTPRequest):
            request: http.HTTPRequest = data
            if request.content and value.lower().strip() in request.content.decode("utf8").lower().strip():
                return True
            if value.lower().strip() in request.pretty_url.lower().strip():
                return True
            if request.headers and value.lower().strip() in str(request.headers).lower().strip():
                return True
        if isinstance(data, http.HTTPResponse):
            response: http.HTTPResponse = data
            if response.content and value.lower().strip() in response.content.decode("utf8").lower().strip():
                return True
            if response.headers and value.lower().strip() in str(response.headers).lower().strip():
                return True
        return False
